import logging
import os
import sys
from typing import Dict

import pandas as pd

from otoole.preprocess.excel_to_osemosys import read_config

logger = logging.getLogger()


def check_parameter(df, config_details, name):
    actual_headers = df.columns
    expected_headers = config_details['indices']
    logger.debug("Expected headers for %s: %s", name, expected_headers)

    if 'REGION' in expected_headers and 'REGION' not in actual_headers:
        logger.info("Added 'REGION' column to %s", name)
        df['REGION'] = 'SIMPLICITY'

    if 'MODEOFOPERATION' in actual_headers:
        df = df.rename(columns={'MODEOFOPERATION': 'MODE_OF_OPERATION'})

    if actual_headers[-1] == 'VALUE':
        logger.info("%s is already in narrow form with headers %s", name, df.columns)
        narrow = df
    else:
        try:
            narrow = pd.melt(df, id_vars=expected_headers[:-1], var_name=expected_headers[-1], value_name='VALUE')
        except IndexError as ex:
            logger.debug(df.columns)
            raise ex

    expected_headers.append('VALUE')
    for column in expected_headers:
        if column not in narrow.columns:
            logger.warning("%s not in header of %s", column, name)

    logger.debug("Final expected headers for %s: %s", name, expected_headers)

    return narrow[expected_headers]


def check_set(df, config_details, name):

    logger.info("Checking set %s", name)
    narrow = df

    return narrow


def check_set_datatype(narrow: pd.DataFrame, config_details: Dict, parameter: str) -> pd.DataFrame:
    datatype = config_details[parameter]['dtype']
    logger.debug('Columns for set %s are: %s', parameter, narrow.columns)
    if narrow.iloc[:, 0].dtype != datatype:
        logger.warning("dtype does not match %s for set %s", datatype, parameter)
    return narrow


def check_datatypes(narrow: pd.DataFrame, config_details: Dict, parameter: str) -> pd.DataFrame:
    """
    """
    logger.info("Checking datatypes for %s", parameter)
    dtypes = {}

    for column in narrow.columns:
        if column == 'VALUE':
            datatype = config_details[parameter]['dtype']
            dtypes['VALUE'] = datatype
        else:
            datatype = config_details[column]['dtype']
            dtypes[column] = datatype
        if narrow[column].dtype != datatype:
            logger.warning("dtype of column %s does not match %s for parameter %s", column, datatype, parameter)
            if datatype == 'int':
                narrow[column] = narrow[column].apply(cast_to_int)
    return narrow.astype(dtypes)


def cast_to_int(value):
    return int(float(value))


def main(output_folder, narrow_folder):
    """Read in a folder of irregular wide-format files and write as narrow csvs
    """
    config = read_config()

    for parameter, details in config.items():
        logger.info("Looking for %s", parameter)
        config_details = config[parameter]

        filepath = os.path.join(output_folder, parameter + '.csv')

        try:
            df = pd.read_csv(filepath)
        except pd.errors.EmptyDataError:
            logger.error("No data found in file for %s", parameter)
            expected_columns = config_details['indices']
            default_columns = expected_columns + ['VALUE']
            df = pd.DataFrame(columns=default_columns)

        entity_type = config[parameter]['type']

        if entity_type == 'param':
            narrow = check_parameter(df, config_details, parameter)
            if not narrow.empty:
                narrow_checked = check_datatypes(narrow, config, parameter)
            else:
                narrow_checked = narrow
        elif entity_type == 'set':
            narrow = check_set(df, config_details, parameter)
            if not narrow.empty:
                narrow_checked = check_set_datatype(narrow, config, parameter)
            else:
                narrow_checked = narrow

        filepath = os.path.join(narrow_folder, 'data', parameter + '.csv')
        with open(filepath, 'w') as csvfile:
            logger.info("Writing %s rows into narrow file for %s", narrow_checked.shape[0], parameter)
            narrow_checked.to_csv(csvfile, index=False)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    output_folder = sys.argv[1]
    narrow_folder = sys.argv[2]
    main(output_folder, narrow_folder)



"""Converts an OSeMOSYS solution file from CPLEX, CBC or GLPK into CBC or CSV format

"""
import argparse
from typing import List, Tuple


class ConvertLine(object):
    """Abstract class which defines the interface to the family of convertors

    Inherit this class and implement the ``_do_it()`` method to produce the
    data to be written out into a new format

    Example
    -------
    >>> cplex_line = "AnnualCost	REGION	CDBACKSTOP	1.0	0.0	137958.8400384134"
    >>> convertor = RegionTechnology()
    >>> convertor.convert()
    VariableName(REGION,TECHCODE01,2015)       42.69         0\\n
    VariableName(REGION,TECHCODE01,2017)       137958.84         0\\n
    """

    def __init__(self, data: List, start_year: int, end_year: int, output_format='cbc'):
        self.data = data
        self.start_year = start_year
        self.end_year = end_year
        self.output_format = output_format

    def _do_it(self) -> Tuple:
        raise NotImplementedError()

    def convert(self) -> List[str]:
        if self.output_format == 'cbc':
            convert = self.convert_cbc()
        elif self.output_format == 'csv':
            convert = self.convert_csv()
        return convert

    def convert_csv(self) -> List[str]:
        """Format the data for writing to a csv file
        """
        data = []
        variable, dimensions, values = self._do_it()

        for index, value in enumerate(values):

            year = self.start_year + index
            if (value not in ["0.0", "0", ""]) and (year <= self.end_year):

                try:
                    value = float(value)
                except ValueError:
                    value = 0

                full_dims = ",".join(dimensions + (str(year),))

                formatted_data = '{0},"{1}",{2}\n'.format(
                    variable,
                    full_dims,
                    value
                    )

                data.append(formatted_data)

        return data

    def convert_cbc(self) -> List[str]:
        """Format the data for writing to a CBC file
        """
        cbc_data = []
        variable, dimensions, values = self._do_it()

        for index, value in enumerate(values):

            year = self.start_year + index
            if (value not in ["0.0", "0", ""]) and (year <= self.end_year):

                try:
                    value = float(value)
                except ValueError:
                    value = 0

                full_dims = ",".join(dimensions + (str(year),))

                formatted_data = "0 {0}({1}) {2} 0\n".format(
                    variable,
                    full_dims,
                    value
                    )

                cbc_data.append(formatted_data)

        return cbc_data


class RegionTimeSliceTechnologyMode(ConvertLine):

    def _do_it(self) -> Tuple:
        """Produces output indexed by Region, Timeslice, Tech and Mode

        ``0 VariableName(REGION,SD1D,TECHCODE01,2,2015) 42.69 0\n``

        """
        variable = self.data[0]
        region = self.data[1]
        timeslice = self.data[2]
        technology = self.data[3]
        mode = self.data[4]
        values = self.data[5:]

        dimensions = (region, timeslice, technology, mode)

        return (variable, dimensions, values)


class RegionTechnology(ConvertLine):

    def _do_it(self) -> Tuple:
        """Produces output indexed by dimensions Region and Technology

        ``0 VariableName(REGION,TECHCODE01,2015) 42.69 0\n``

        """
        variable = self.data[0]
        region = self.data[1]
        technology = self.data[2]

        dimensions = (region, technology)

        values = self.data[3:]

        return (variable, dimensions, values)


def process_line(line: str,
                 start_year: int,
                 end_year: int,
                 output_format: str) -> List[str]:
    """Processes an individual line in a CPLEX file

    A different ConvertLine implementation is chosen depending upon the
    variable name

    Arguments
    ---------
    line: str
    start_year: int
    end_year: int
    output_format: str
        The file format required - either ``csv`` or ``cbc``
    """
    row_as_list = line.split('\t')
    variable = row_as_list[0]
    if variable in ['NewCapacity',
                    'TotalCapacityAnnual',
                    'CapitalInvestment',
                    'AnnualFixedOperatingCost',
                    'AnnualVariableOperatingCost']:
        convertor = RegionTechnology(row_as_list, start_year, end_year, output_format).convert()
    elif variable in ['RateOfActivity']:
        convertor = RegionTimeSliceTechnologyMode(row_as_list, start_year, end_year, output_format).convert()
    else:
        convertor = []

    return convertor


def convert_cplex_file(cplex_filename: str, output_filename: str,
                       start_year=2015, end_year=2070,
                       output_format='cbc'):
    """Converts a CPLEX solution file into that of the CBC solution file

    Arguments
    ---------
    cplex_filename : str
        Path to the transformed CPLEX solution file
    output_filename : str
        Path for the processed data to be written to
    """

    with open(output_filename, 'w') as cbc_file:
        with open(cplex_filename, 'r') as cplex_file:
            for linenum, line in enumerate(cplex_file):
                try:
                    convertor = process_line(line, start_year, end_year, output_format)
                    if convertor:
                        cbc_file.writelines(convertor)
                except ValueError:
                    msg = "Error caused at line {}: {}"
                    raise ValueError(msg.format(linenum, line))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Convert OSeMOSYS CPLEX files into different formats")
    parser.add_argument("cplex_file",
                        help="The filepath of the OSeMOSYS cplex output file")
    parser.add_argument("output_file",
                        help="The filepath of the converted file that will be written")
    parser.add_argument("-s", "--start_year", type=int, default=2015,
                        help="Output only the results from this year onwards")
    parser.add_argument("-e", "--end_year", type=int, default=2070,
                        help="Output only the results upto and including this year")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--csv", action="store_true",
                       help="Output file in comma-separated-values format")
    group.add_argument("--cbc", action="store_true",
                       help="Output file in CBC format, (default option)")

    args = parser.parse_args()

    if args.csv:
        output_format = 'csv'
    else:
        output_format = 'cbc'

    convert_cplex_file(args.cplex_file, args.output_file,
                       args.start_year, args.end_year,
                       output_format)



"""Provides a command line interface to the ``otoole`` package

Example
-------
>>> otoole --help
usage: otoole [-h] [-v] {prep,cplex} ...

otoole: Python toolkit of OSeMOSYS users

positional arguments:
  {prep,cplex}
    prep        Prepare an OSeMOSYS datafile
    cplex       Process a CPLEX solution file

optional arguments:
  -h, --help    show this help message and exit
  -v            Enable debug mode

"""
import argparse
import logging
import sys

from otoole.preprocess import create_datafile, create_datapackage, generate_csv_from_excel
from otoole.results.convert import convert_cplex_file


def excel2csv(args):
    generate_csv_from_excel(args.workbook, args.output_folder)


def csv2datapackage(args):
    create_datapackage(args.csv_folder, args.datapackage)


def cplex2cbc(args):
    convert_cplex_file(args.cplex_file, args.output_file, args.start_year,
                       args.end_year, args.output_format)


def datapackage2datafile(args):
    create_datafile(args.datapackage, args.datafile)


def get_parser():
    parser = argparse.ArgumentParser(description="otoole: Python toolkit of OSeMOSYS users")

    parser.add_argument('-v', help='Enable debug mode', action='store_true')

    subparsers = parser.add_subparsers()

    # Parser for pre-processing related commands
    prep_parser = subparsers.add_parser('prep', help='Prepare an OSeMOSYS datafile')
    prep_subparsers = prep_parser.add_subparsers()

    excel_parser = prep_subparsers.add_parser('excel', help='Convert from an Excel workbook')
    excel_parser.add_argument('workbook', help='Path to the Excel workbook')
    excel_parser.add_argument('output_folder', help='Folder to which to write csv files')
    excel_parser.set_defaults(func=excel2csv)

    datapackage_parser = prep_subparsers.add_parser('datapackage',
                                                    help='Convert a folder of csv file to a datapackage')
    datapackage_parser.add_argument('csv_folder', help='Path to folder containing csv files')
    datapackage_parser.add_argument('datapackage', help='Path to destination for datapackage')
    datapackage_parser.set_defaults(func=csv2datapackage)

    datafile_parser = prep_subparsers.add_parser('datafile',
                                                 help='Convert an OSeMOSYS datapackage to a datafile')
    datafile_parser.add_argument('datapackage', help='Path to destination for datapackage')
    datafile_parser.add_argument('datafile', help='Path to datafile')
    datafile_parser.set_defaults(func=datapackage2datafile)

    # Parser for the CPLEX related commands
    cplex_parser = subparsers.add_parser('cplex',
                                         help='Process a CPLEX solution file')

    cplex_parser.add_argument("cplex_file",
                              help="The filepath of the OSeMOSYS cplex output file")
    cplex_parser.add_argument("output_file",
                              help="The filepath of the converted file that will be written")
    cplex_parser.add_argument("-s", "--start_year", type=int, default=2015,
                              help="Output only the results from this year onwards")
    cplex_parser.add_argument("-e", "--end_year", type=int, default=2070,
                              help="Output only the results upto and including this year")
    cplex_parser.add_argument('output_format', choices=['csv', 'cbc'], default='cbc')
    cplex_parser.set_defaults(func=cplex2cbc)
    return parser


def main():

    logging.basicConfig(filename='myapp.log', level=logging.DEBUG, filemode='w')
    logging.info('Started')
    parser = get_parser()
    args = parser.parse_args(sys.argv[1:])

    if args.v:
        logging.basicConfig(filename='myapp.log', level=logging.DEBUG, filemode='a')

    if 'func' in args:
        args.func(args)
    else:
        parser.print_help()
    logging.info('Finished')



"""Creates a datapackage from a collection of CSV files of OSeMOSYS input data

- Uses Frictionless Data datapackage concept to build a JSON schema of the dataset
- Enforces relations between sets and indices in parameter files
"""

import logging
import os
import sys

from datapackage import Package

from otoole.preprocess.excel_to_osemosys import read_config
from otoole.preprocess.longify_data import main as longify

logger = logging.getLogger()


def generate_package(path_to_package):
    """Creates a datapackage in folder ``path_to_package``

    [{'fields': 'REGION', 'reference': {'resource': 'REGION', 'fields': 'VALUE'}}]
    """

    datapath = os.path.join(path_to_package)
    package = Package(base_path=datapath)

    package.infer('data/*.csv')

    package.descriptor['licenses'] = [{"name": "CC-BY-4.0",
                                       "path": "https://creativecommons.org/licenses/by/4.0/",
                                       "title": "Creative Commons Attribution 4.0"}]

    package.descriptor['title'] = 'The OSeMOSYS Simplicity Example Model'

    package.descriptor['name'] = 'osemosys_model_simplicity'

    package.descriptor['contributors'] = [{"title": "Will Usher",
                                           "email": "wusher@kth.se",
                                           "path": "http://www.kth.se/wusher",
                                           "role": "author"
                                           }]

    package.commit()

    config = read_config()

    new_resources = []
    for resource in package.resources:

        descriptor = resource.descriptor

        name = resource.name
        if config[name]['type'] == 'param':

            indices = config[name]['indices']
            logger.debug("Indices of %s are %s", name, indices)

            foreign_keys = []
            for index in indices:
                key = {'fields': index, 'reference': {'resource': index, 'fields': 'VALUE'}}
                foreign_keys.append(key)

            descriptor['schema']['foreignKeys'] = foreign_keys
            descriptor['schema']['primaryKey'] = indices
            descriptor['schema']['missingValues'] = [str(config[name]['default'])]

        new_resources.append(descriptor)

    package.descriptor['resources'] = new_resources
    package.commit()

    filepath = os.path.join(path_to_package, 'datapackage.json')
    package.save(filepath)


def validate_contents(path_to_package):

    filepath = os.path.join(path_to_package)
    package = Package(filepath)

    for resource in package.resources:
        try:
            if resource.check_relations():
                logger.info("%s is valid", resource.name)
        except KeyError as ex:
            logger.warning("Validation error in %s: %s", resource.name, str(ex))


def main(wide_folder, narrow_folder):
    longify(wide_folder, narrow_folder)
    generate_package(narrow_folder)
    absolute_path = os.path.join(narrow_folder, 'datapackage.json')
    validate_contents(absolute_path)


if __name__ == '__main__':

    logging.basicConfig(level=logging.DEBUG)
    wide_folder = sys.argv[1]
    narrow_folder = sys.argv[2]
    main(wide_folder, narrow_folder)



from .excel_to_osemosys import generate_csv_from_excel
from .create_datapackage import main as create_datapackage
from .narrow_to_datafile import main as create_datafile

__all__ = ['generate_csv_from_excel', 'create_datapackage', 'create_datafile']



import io
import logging
import sys

import pandas as pd
from datapackage import Package

logger = logging.getLogger(__name__)


def write_parameter(filepath: io.TextIOBase, df: pd.DataFrame, parameter_name, default):
    """

    Arguments
    ---------
    filepath : StreamIO
    df : pandas.DataFrame
    parameter_name : str
    default : int
    """
    filepath.write('param default {} : {} :=\n'.format(default, parameter_name))
    df.to_csv(path_or_buf=filepath, sep=" ", header=False, index=False)
    filepath.write(';\n')
    return filepath


def write_set(filepath: io.TextIOBase, df: pd.DataFrame, set_name):
    """

    Arguments
    ---------
    filepath : StreamIO
    df : pandas.DataFrame
    parameter_name : str
    """
    filepath.write('set {} :=\n'.format(set_name))
    df.to_csv(path_or_buf=filepath, sep=" ", header=False, index=False)
    filepath.write(';\n')
    return filepath


def read_narrow_csv(filepath):

    df = pd.read_csv(filepath)
    return df


def main(datapackage, datafilepath):

    package = Package(datapackage)

    with open(datafilepath, 'w') as filepath:

        for resource in package.resources:
            if resource.check_relations():
                logger.info("%s is valid", resource.name)

                data = resource.read()

                if data:
                    headers = resource.headers
                else:
                    fields = resource.descriptor['schema']['fields']
                    headers = [x['name'] for x in fields]
                df = pd.DataFrame(data, columns=headers)

                default_value = resource.descriptor['schema']['missingValues'][0]
                if len(headers) > 1:
                    write_parameter(filepath, df, resource.name, default=default_value)
                else:
                    write_set(filepath, df, resource.name)
            else:
                logger.warning("Validation error in %s", resource.name)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    path_to_datapackage = sys.argv[1]
    path_to_datafile = sys.argv[2]
    main(path_to_datapackage, path_to_datafile)



#!/usr/bin/env python
# coding: utf-8
"""Extract data from spreadsheets and write to an OSeMOSYS datafile

Reads and writes the following OSeMOSYS parameters and sets to and from
files on disk.

Notes
=====

Sets
~~~~
set YEAR;
set TECHNOLOGY;
set TIMESLICE;
set FUEL;
set EMISSION;
set MODE_OF_OPERATION;
set REGION;
set SEASON;
set DAYTYPE;
set DAILYTIMEBRACKET;
set FLEXIBLEDEMANDTYPE;
set STORAGE;

All sets are written in a CSV file in one column of values with no header.
For example, the CSV file for set YEAR::

    2015
    2016
    2017
    2018
    2019
    2020

Sets are written into the OSeMOSYS data file using the following syntax::

    set YEAR := 2014 2015 2016 2017 2018 2019 2020 ;

Parameters
~~~~~~~~~~

In general, parameters can be written into CSV files in narrow or wide
formats. In narrow format, the CSV file should look as follows::

    TIMESLICE,YEAR,VALUE
    ID,2014,0.1667
    IN,2014,0.0833
    SD,2014,0.1667
    SN,2014,0.0833
    WD,2014,0.3333
    WN,2014,0.1667

In wide format, the final index is transposed::

    TIMESLICE,2014,2015,...
    ID,0.1667,0.1667
    IN,0.0833,0.0833
    SD,0.1667,0.1667
    SN,0.0833,0.0833
    WD,0.3333,0.3333
    WN,0.1667,0.1667

Wide format is a bit nicer to use with spreadsheets,
as it allows you to more easily plot graphs of values, but narrow
format is a more flexible and easily manipulated data format.

Writing parameteters:

1-dimensional e.g. DiscountRate{r in REGION}

2-dimension e.g. YearSplit{l in TIMESLICE, y in YEAR}

n-dimensional e.g. DaysInDayType{ls in SEASON, ld in DAYTYPE, y in YEAR}



Global
------
param YearSplit{l in TIMESLICE, y in YEAR};
param DiscountRate{r in REGION};
param DaySplit{lh in DAILYTIMEBRACKET, y in YEAR};
param Conversionls{l in TIMESLICE, ls in SEASON};
param Conversionld{l in TIMESLICE, ld in DAYTYPE};
param Conversionlh{l in TIMESLICE, lh in DAILYTIMEBRACKET};
param DaysInDayType{ls in SEASON, ld in DAYTYPE, y in YEAR};
param TradeRoute{r in REGION, rr in REGION, f in FUEL, y in YEAR};
param DepreciationMethod{r in REGION};

Demands
-------
param SpecifiedAnnualDemand{r in REGION, f in FUEL, y in YEAR};
param SpecifiedDemandProfile{r in REGION, f in FUEL, l in TIMESLICE, y in YEAR};
param AccumulatedAnnualDemand{r in REGION, f in FUEL, y in YEAR};

Performance
-----------

param CapacityToActivityUnit{r in REGION, t in TECHNOLOGY};
param TechWithCapacityNeededToMeetPeakTS{r in REGION, t in TECHNOLOGY};
param CapacityFactor{r in REGION, t in TECHNOLOGY, l in TIMESLICE, y in YEAR};
param AvailabilityFactor{r in REGION, t in TECHNOLOGY, y in YEAR};
param OperationalLife{r in REGION, t in TECHNOLOGY};
param ResidualCapacity{r in REGION, t in TECHNOLOGY, y in YEAR};
param InputActivityRatio{r in REGION, t in TECHNOLOGY, f in FUEL, m in MODE_OF_OPERATION, y in YEAR};
param OutputActivityRatio{r in REGION, t in TECHNOLOGY, f in FUEL, m in MODE_OF_OPERATION, y in YEAR};

Technology Costs
----------------
param CapitalCost{r in REGION, t in TECHNOLOGY, y in YEAR};
param VariableCost{r in REGION, t in TECHNOLOGY, m in MODE_OF_OPERATION, y in YEAR};
param FixedCost{r in REGION, t in TECHNOLOGY, y in YEAR};

Storage
-------
param TechnologyToStorage{r in REGION, t in TECHNOLOGY, s in STORAGE, m in MODE_OF_OPERATION};
param TechnologyFromStorage{r in REGION, t in TECHNOLOGY, s in STORAGE, m in MODE_OF_OPERATION};
param StorageLevelStart{r in REGION, s in STORAGE};
param StorageMaxChargeRate{r in REGION, s in STORAGE};
param StorageMaxDischargeRate{r in REGION, s in STORAGE};
param MinStorageCharge{r in REGION, s in STORAGE, y in YEAR};
param OperationalLifeStorage{r in REGION, s in STORAGE};
param CapitalCostStorage{r in REGION, s in STORAGE, y in YEAR};
param ResidualStorageCapacity{r in REGION, s in STORAGE, y in YEAR};

Capacity Constraints
--------------------
param CapacityOfOneTechnologyUnit{r in REGION, t in TECHNOLOGY, y in YEAR};
param TotalAnnualMaxCapacity{r in REGION, t in TECHNOLOGY, y in YEAR};
param TotalAnnualMinCapacity{r in REGION, t in TECHNOLOGY, y in YEAR};

Investment Constraints
----------------------
param TotalAnnualMaxCapacityInvestment{r in REGION, t in TECHNOLOGY, y in YEAR};
param TotalAnnualMinCapacityInvestment{r in REGION, t in TECHNOLOGY, y in YEAR};

Activity Constraints
--------------------
param TotalTechnologyAnnualActivityUpperLimit{r in REGION, t in TECHNOLOGY, y in YEAR};
param TotalTechnologyAnnualActivityLowerLimit{r in REGION, t in TECHNOLOGY, y in YEAR};
param TotalTechnologyModelPeriodActivityUpperLimit{r in REGION, t in TECHNOLOGY};
param TotalTechnologyModelPeriodActivityLowerLimit{r in REGION, t in TECHNOLOGY};

Reserve Margin
--------------
param ReserveMarginTagTechnology{r in REGION, t in TECHNOLOGY, y in YEAR};
param ReserveMarginTagFuel{r in REGION, f in FUEL, y in YEAR};
param ReserveMargin{r in REGION, y in YEAR};

RE Generation Target
--------------------
param RETagTechnology{r in REGION, t in TECHNOLOGY, y in YEAR};
param RETagFuel{r in REGION, f in FUEL, y in YEAR};
param REMinProductionTarget{r in REGION, y in YEAR};

Emissions & Penalties
---------------------
param EmissionActivityRatio{r in REGION, t in TECHNOLOGY, e in EMISSION, m in MODE_OF_OPERATION, y in YEAR};
param EmissionsPenalty{r in REGION, e in EMISSION, y in YEAR};
param AnnualExogenousEmission{r in REGION, e in EMISSION, y in YEAR};
param AnnualEmissionLimit{r in REGION, e in EMISSION, y in YEAR};
param ModelPeriodExogenousEmission{r in REGION, e in EMISSION};
param ModelPeriodEmissionLimit{r in REGION, e in EMISSION};


"""
import csv
import logging
import os
from typing import List

import xlrd
from yaml import SafeLoader, load

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources

logger = logging.getLogger(__name__)


def read_config():

    with pkg_resources.open_text('otoole.preprocess', 'config.yaml') as config_file:
        config = load(config_file, Loader=SafeLoader)
    return config


def generate_csv_from_excel(input_workbook, output_folder):
    """Generate a folder of CSV files from a spreadsheet

    Arguments
    ---------
    input_workbook : str
        Path to spreadsheet containing OSeMOSYS data
    output_folder : str
        Path of the folder containing the csv files

    """
    work_book = xlrd.open_workbook(os.path.join(input_workbook))

    _csv_from_excel(work_book, output_folder)
    work_book.release_resources()  # release the workbook-resources
    del work_book


def _csv_from_excel(workbook, output_folder):
    """Creates csv files from all sheets in a workbook

    Arguments
    ---------
    workbook :
    output_folder : str
    """

    logger.debug("Generating CSVs from Excel %s", workbook)
    # Create all the csv files in a new folder
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)  # creates the csv folder

    # Iterate over each sheet in the workbook
    for sheet in workbook.sheets():  # typing: xlrd.book.Sheet

        name = sheet.name
        mod_name = _modify_names([name])

        # Open the sheet name in the xlsx file and write it in csv format]
        filepath = os.path.join(output_folder, mod_name[0] + '.csv')
        with open(filepath, 'w', newline='') as your_csv_file:
            wr = csv.writer(your_csv_file, quoting=csv.QUOTE_NONNUMERIC)

            for rownum in range(sheet.nrows):  # reads each row in the csv file
                row = _cast_to_integer(sheet.row_values(rownum))
                wr.writerow(row)


def _modify_names(sheet_names: List) -> List:
    """I change the name of the sheets in the xlsx file to match with the csv
    actual ones
    """
    modifiedNames = sheet_names.copy()
    for name in modifiedNames:
        if (name == "TotalAnnualMaxCapacityInvestmen"):
            name = "TotalAnnualMaxCapacityInvestment"
        elif (name == "TotalAnnualMinCapacityInvestmen"):
            name = "TotalAnnualMinCapacityInvestment"
        elif (name == "TotalTechnologyAnnualActivityLo"):
            name = "TotalTechnologyAnnualActivityLowerLimit"
        elif (name == "TotalTechnologyAnnualActivityUp"):
            name = "TotalTechnologyAnnualActivityUpperLimit"
        elif (name == "TotalTechnologyModelPeriodActLo"):
            name = "TotalTechnologyModelPeriodActivityLowerLimit"
        elif (name == "TotalTechnologyModelPeriodActUp"):
            name = "TotalTechnologyModelPeriodActivityUpperLimit"
    return modifiedNames


def _cast_to_integer(row):
    """function to convert all float numbers to integers....need to check it!!
    """
    if all(isinstance(n, float) for n in row):
        converted_row = list(map(int, row))
    else:
        converted_row = row
    return converted_row
